# Copyright (C) 2004 Laurent Sansonetti
#
# Tengen is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Tengen is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with Tengen; see the file COPYING.  If not,
# write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

require 'observer'
require 'singleton'
require 'thread'

module Tengen
    class Game
        attr_reader :white_player, :black_player, :current_player,
                    :winner, :score, :start_time, :end_time
        attr_accessor :komi, :boardsize, :handicap
       
        include Observable

        class Engine < GTP
            include Singleton

            def initialize
                super
                @notify_genmove = Mutex.new.lock
                Thread.start do
                    loop do
                        @notify_genmove.lock
                        Tengen.debug("Executing genmove for " \
                                     "#{@genmove_requester}.")
                        @genmove_requester.do_genmove
                    end
                end
            end
        
            def request_genmove(player)
                @genmove_requester = player
                Tengen.debug("Requesting genmove for #{player}.")
                @notify_genmove.unlock
            end

            def reset!
                @resetting = true
                Tengen.debug("Resetting engine...")
                clear_board
                Tengen.debug("Reset OK.")
                @resetting = false
            end

            def resetting?; @resetting; end
        end

        def initialize(komi=6.5, boardsize=19, handicap=nil)
            @komi = komi 
            @boardsize = boardsize
            @handicap = handicap
        end
        
        def white_player=(player)
            raise "Someone already playing" unless @white_player.nil?
            @white_player = player
            @white_player.enter(self, Player::WHITE)
        end
        
        def black_player=(player)
            raise "Someone already playing" unless @black_player.nil?
            @black_player = player
            @black_player.enter(self, Player::BLACK)
        end
   
        def first_player
            # black plays first unless there is an handicap
            (@handicap.nil?) ? @black_player : @white_player
        end
   
        def start
            raise "Black player is missing" if @black_player.nil?
            raise "White player is missing" if @white_player.nil?

            @start_time = Time.now
            Engine.instance.reset!
            Engine.instance.boardsize = @boardsize
            Engine.instance.komi = @komi
            Engine.instance.fixed_handicap(@handicap) unless @handicap.nil?

            @current_player = first_player
            @current_player.genmove if @current_player.computer?
        end
    
        def started?
            @start_time != nil
        end
    
        def finished?
            @end_time != nil
        end
    
        def played(player, action)
            raise "Unknown player" unless @white_player == player or
                                          @black_player == player
            raise "Invalid action" unless action.kind == Action::MOVE or
                                          action.kind == Action::PASS

            if @last_move and @last_move.pass? and action.pass?
                guess_score
                terminate
            elsif not Engine.instance.resetting?
                @last_move = action 
                @current_player = other(player)
                changed and notify_observers(self, player, action)
            end
        end
    
        def resigned(player)
            @winner = other(player)
            @score = @winner.color[0..0].upcase + "+R" 
            terminate
        end
   
        def when_finished(&on_terminate_cb)
            raise if on_terminate_cb.nil?
            @on_terminate_cb ||= [] 
            @on_terminate_cb << on_terminate_cb
        end
   
        def other(player)
            case player
                when @white_player then @black_player
                when @black_player then @white_player
                else
                    raise "Unknown player"
            end
        end

        def undo
            2.times { Engine.instance.undo }
            changed and notify_observers(self, @current_player, 
                                         Action.new_undo)
        end

        #######
        private
        #######
  
        def guess_score
            @score = Engine.instance.final_score
            @winner = case @score[0]
                when ?W then @white_player
                when ?B then @black_player
            end
        end
        
        def terminate
            @end_time = Time.now
            @on_terminate_cb.each { |x| x.call(self) } if @on_terminate_cb 
        end
    end
end
